#! /usr/bin/env python
# -*- coding: utf-8 -*-
## vim: fileencoding=utf-8

# Copyright (c) 2009 Nicholas Allen (nick.allen@onlinehome.de)
#                    Adeodato Simó (dato@net.com.org.es)
# Licensed under the terms of the MIT license.

"""Send commit emails for Bazaar branches."""

import os
import sys
import socket
import time
import optparse
import StringIO
import cStringIO

from bzrlib import revision as _mod_revision
from bzrlib import version_info as _bzrlib_version_info
from bzrlib.branch import Branch
from bzrlib.bzrdir import BzrDir
from bzrlib.diff import show_diff_trees
from bzrlib.errors import (NotBranchError, NoSuchRevision, LockFailed, BzrError)
from bzrlib.errors import SMTPError
from bzrlib.log import log_formatter, show_log
from bzrlib.osutils import format_date
from bzrlib.plugin import load_plugins

from configuration import Configuration, SendEmailError
###

def main():

    try:
        load_plugins()
    except BzrError, e:
        print >>sys.stderr, "Could not load plugins: %s" % (e)
        sys.exit(1)

    global options
    options, arguments = parse_options()

    if not arguments:
        print >>sys.stderr, 'E: no directories given.'
        sys.exit(1)
        
    # Strip trailing '/' from any paths provided.
    directories = [path if path[-1] != '/' else path[:-1] for path in arguments]
        
    global config
    config = Configuration(options.config_file)
    
    global transports
    transports = []
    
    # Load state file so we can detect new commits, branches and removed branches
    # since the last run. last_emailed_revisions is a map from a branch's relative
    # path to its last emailed revision id.
    
    if os.path.exists(config.state_file):
        state_file = open(config.state_file, 'r')
        last_emailed_revisions = eval(state_file.read())
        state_file.close()
    else:
        last_emailed_revisions = dict()
        
    # Detect what branches exist now and compare that with what we had before
    # to find newly added and removed branches.
    
    branches = dict()  # Map from branch relative path to Bazaar branch object.
    
    for directory in directories:
        detect_branches(branches, directory)  
        
    # Send emails for any branch that has been removed.
    send_removed_branch_emails(last_emailed_revisions, branches)
    
    # Send commit emails for any new revisions in branches that previously existed.
    for relative_path, branch in branches.items():
        
        # Pull from parent if this is desired.
        if config.pull_from_parent:
            try:
                pull_from_parent(branch)
            except BzrError, e:
                print >>sys.stderr, \
                        'Could not pull from parent for branch %s: %s' % (branch_emailer.path, e) 
        
        # Update working tree if branch has one and this is desired.
        if config.update_working_tree:
            try:
                if branch.bzrdir.has_workingtree():
                    if options.verbose:
                        print >>sys.stdout, "Udpating working tree for branch %s" % (branch_emailer.path)
                        
                    working_tree = branch.bzrdir.open_workingtree()  
                    working_tree.update()                         
            except BzrError, e:
                print >>sys.stderr, \
                        'Could not update working tree for branch %s: %s' % (branch_emailer.path, e) 
            
        # Finally send emails for this branch.
        if last_emailed_revisions.has_key(relative_path):
            send_commit_emails_for_branch(branch, relative_path, last_emailed_revisions)
            
    # Send emails for any new branches that have been detected.
    send_new_branch_emails(last_emailed_revisions, branches)
        
    sys.exit(0)
      
def detect_branches(branches, directory, relative_path = ""):
    # Recursively scans a directory for branches and adds them to the list
    # of branches provided.    
    try:
        branch = Branch.open(directory)
        branches[relative_path] = branch
    except BzrError,e:
        # Not a branch so scan sub directories to see if they contain a branch.
        for file_name in os.listdir(directory):
            file_path = os.path.join(directory, file_name)
            if os.path.isdir(file_path):
                if relative_path != "":
                    relative_file_path = relative_path + "/" + file_name
                else:
                    relative_file_path = file_name
                    
                detect_branches(branches, file_path, relative_file_path)       
                
                
def send_removed_branch_emails(last_emailed_revisions, branches):
    # Sends an email for each branch that no longer exists since the last time
    # we were run.            
    
    if config.removed_branch_email_config is None:
        # Don't send emails as this option has not been configured.
        return
        
    for relative_path in last_emailed_revisions.keys():
        if not branches.has_key(relative_path):
            send_removed_branch_email(relative_path, last_emailed_revisions)
            
def send_removed_branch_email(relative_path, last_emailed_revisions):
    # Sends an email for a branch that has been removed.
    if options.verbose:
        print >>sys.stdout, "Sending removed branch email for branch at %s" % relative_path
        
    message = compose_removed_branch_email(relative_path)
    
    try:
        config.send_email(message)
        del last_emailed_revisions[relative_path]
        save_last_emailed_revisions(last_emailed_revisions)
    except SendEmailError, e:
        print >>sys.stderr, \
                        'Could not send removed branch email %s' % (e)
                        
def compose_removed_branch_email(relative_path):
    template_data = {
        'relative_path' : relative_path
    }
    
    return config.removed_branch_email_config.compose_email(template_data)   
   
   
def send_new_branch_emails(last_emailed_revisions, branches):
    # Sends an email for each branch that did not exist previously but has
    # now been detected.
    
    if config.new_branch_email_config is None:
        return
        
    for relative_path, branch in branches.items():
        if not last_emailed_revisions.has_key(relative_path):
            send_new_branch_email(branch, relative_path, last_emailed_revisions)

            
def send_new_branch_email(branch, relative_path, last_emailed_revisions):
    # Sends an email for a newly detected branch.
        
    if options.verbose:
        print >>sys.stdout, "Sending new branch email for branch %s" % relative_path
        
    try:
        message = compose_new_branch_email(branch, relative_path)
        config.send_email(message)
        last_emailed_revisions[relative_path] = branch.last_revision()
        save_last_emailed_revisions(last_emailed_revisions)
    except (BzrError, SendEmailError), e:
        print >>sys.stderr, \
                        'Could not send new branch email %s' % (e)


def compose_new_branch_email(branch, relative_path):
    template_data = {
        'branch' : branch,
        'relative_path' : relative_path
    }
    
    return config.new_branch_email_config.compose_email(template_data)
        
                
def send_commit_emails_for_branch(branch, relative_path, last_emailed_revisions):        
    # Sends a commit email for each new commit after last_emailed_revision.
    
    if config.commit_email_config is None:
            return
            
    last_emailed_revision = last_emailed_revisions[relative_path]
            
    for revid in revisions_to_send(branch, last_emailed_revision):
        if options.verbose:
            print >>sys.stdout, "Sending commit email for revision %s in branch %s" % (revid, relative_path)
    
        try:
            message = compose_commit_email(branch, revid, relative_path)
            config.send_email(message)
            last_emailed_revisions[relative_path] = revid
            save_last_emailed_revisions(last_emailed_revisions)
        except (BzrError, SendEmailError), e:
            print >>sys.stderr, \
                        'Could not send revision %s: %s' % (revision, e)
            return

    
def revisions_to_send(branch, last_emailed_revision):
    # Determines which revisions we have not sent email messages for.        
    revision_history = branch.revision_history()
    
    assert not last_emailed_revision is None
    
    if last_emailed_revision == 'null:':
        # This was a newly detected branch that did not have any revisions
        # when we first detected it.
        return revision_history

    if last_emailed_revision not in revision_history:
        # The tip we last saw disappeared, either by uncommit or by having
        # been merged by a new mainline revision. So, we use as last_mailed
        # revision the first mainline ancestor of the disappeared one.
        revision = last_emailed_revision.encode('utf-8')
        repo = branch.repository
        
        while True:
            try:
                revobj = repo.get_revision(revision)
            except NoSuchRevision:
                print >>sys.stderr, (
                    'E: %s does not exist in the repository for %s\n'
                    'E: setting %s to the last available revision' % (
                        last_mailed, self.path, self._option_name))
                self._config.set_user_option(
                        self._option_name, revision_history[-1])
                return []

            if revobj.parent_ids:
                revision = revobj.parent_ids[0]
                if revision in revision_history:
                    last_emailed_revision = revision
                    break
            else:
                return revision_history

    index = revision_history.index(last_emailed_revision)
    return revision_history[index + 1:]
    

def compose_commit_email(branch, revid, relative_path):
    # Composes the email that is sent for each committed revision.
    ### This code is based/stolen from that in bzr-email/emailer.py:
    ### Copyright (C) 2005, 2006, 2007 Canonical Ltd.
    ### Licensed under the terms of GPLv2 or later.
    revision_number = branch.revision_id_to_revno(revid) or None
    revision = branch.repository.get_revision(revid)
    log = StringIO.StringIO()

    lf = log_formatter('long', to_file=log)
    show_log(branch, lf, start_revision=revision_number, end_revision=revision_number,
            verbose=True, direction='forward')
            
    changes = branch.get_revision_delta(revision_number)
    
    template_data = {
        'branch' : branch,
        'relative_path' : relative_path,
        'revision_number' : revision_number,
        'revision_id' : revid,
        'revision' : revision,
        'summary' : revision.get_summary,
        'changes' : changes
    }
            
    msg = config.commit_email_config.compose_email(template_data)
    msg['Date'] = format_date(revision.timestamp, revision.timezone,
            date_fmt='%a, %d %b %Y %H:%M:%S')
            
    # Attach the diffs if less than the maximum line limit.
    if config.diff_line_limit != 0:
        diff = cStringIO.StringIO()
        diff_name = str(config.diff_attachment_name(template_data))

        revid_new = revision.revision_id

        branch.repository.lock_read()
        
        try:
            if revision.parent_ids:
                revid_old = revision.parent_ids[0]
                tree_new, tree_old = branch.repository.revision_trees(
                        (revid_new, revid_old))
            else:
                # revision_trees() doesn't allow None or 'null:' to be passed
                # as a revision. So we need to call revision_tree() twice.
                revid_old = _mod_revision.NULL_REVISION
                tree_new = branch.repository.revision_tree(revid_new)
                tree_old = branch.repository.revision_tree(revid_old)
        finally:
            branch.repository.unlock()

        show_diff_trees(tree_old, tree_new, diff)
        numlines = diff.getvalue().count('\n') + 1
        if numlines <= config.diff_line_limit or config.diff_line_limit < 0:
            diff = diff.getvalue()
        else:
            diff = ("Diff too large for email (%d lines, the limit is %d).\n" %
                    (numlines, config.diff_line_limit))

        msg.add_inline_attachment(diff, diff_name)

    return msg                
                         


def pull_from_parent(branch):
    parent_location = branch.get_parent()
            	
    if not parent_location is None:
        if options.verbose:
            print >>sys.stdout, "Pulling from parent location %s" % (parent_location)
            
        parent_branch = Branch.open(parent_location, possible_transports=transports)                        
        branch.pull(parent_branch)
        

def save_last_emailed_revisions(last_emailed_revisions):
    # Writes the last-emailed-revisions.txt file.

    state_file = open(config.state_file, 'w')
    state_file.write(str(last_emailed_revisions))
    state_file.close()
    

def parse_options():
    p = optparse.OptionParser(usage='%prog -c CONFIG_FILE [options] DIR [ DIR ... ]')

    p.add_option('-c', '--config', action='store', dest='config_file',
            help='config file defining settings to use')

    p.add_option('-v', '--verbose', action='store_true', dest='verbose', 
            help='display verbose messages about what the program is doing')
            
    p.set_defaults(bg_fork=True, logfile='/dev/null')

    options, args = p.parse_args()

    return options, args

###

if __name__ == '__main__':
    main()
