# -*- coding: utf-8 -*-
from email import Utils
import smtplib
import ConfigParser, os, sys, socket
from bzrlib.email_message import EmailMessage
from Cheetah.Template import Template

class Configuration:
    # Defines configuration options for the email notifier.
    
    def __init__(self, config_file):
        self._config_file = config_file
        self.commit_email_config = None
        self.new_branch_email_config = None
        self.removed_branch_email_config = None
        self._connection = None
        self._load()
        
        
    def _load(self):
        # Loads settings from the configuration file.
        
        last_modified_time = os.path.getmtime(self._config_file)
        config = ConfigParser.ConfigParser()
        config.read(self._config_file)
        self.state_file = get_option(config, 'DEFAULT', 'state-file', 'last-emailed-revisions.txt')
        self._smtp_server = config.get('smtp', 'server')
        self._smtp_username = get_option(config, 'smtp', 'user', None)
        self._smtp_password = get_option(config, 'smtp', 'password', None)
        self.pull_from_parent = get_bool_option(config, 'DEFAULT', 'pull-from-parent', False)
        self.update_working_tree = get_bool_option(config, 'DEFAULT', 'update-working-tree', False)
        
        # Parse commit emails settings.
        
        if config.has_section('commits'):
            if self.commit_email_config is None:
                self.commit_email_config = EmailConfiguration(self)
                
            self.commit_email_config.load(config, 'commits')
            self.diff_line_limit = get_int_option(config, 'commits', 'diff-line-limit', 100)
            self.diff_attachment_name = Template.compile(source=get_option(config, 'commits', 'diff-attachment-name', 'r${revision_number}.diff'), baseclass=dict)
        else:
            self.commit_email_config = None
            print >>sys.stderr, "No 'commits' section in configuration - commit emails won't be sent"
            
        # Parse new branch email settings.
            
        if config.has_section('new-branches'):
            if self.new_branch_email_config is None:
                self.new_branch_email_config = EmailConfiguration(self)
                
            self.new_branch_email_config.load(config, 'new-branches')
        else:
            self.new_branch_email_config = None
            print >>sys.stderr, "No 'new-branches' section in configuration - emails won't be sent when new branches are created"
        
        # Parse removed branch email settings.
        
        if config.has_section('removed-branches'):
            if self.removed_branch_email_config is None:
                self.removed_branch_email_config = EmailConfiguration(self)
                
            self.removed_branch_email_config.load(config, 'removed-branches')
        else:
            self.removed_branch_email_config = None
            print >>sys.stderr, "No 'removed-branches' section in configuration - emails won't be sent when branches are deleted"
        
        if config.has_section('template-settings'):
            self.template_settings = config.items('template-settings')
        else:
            self.template_settings = {}
            
        self._last_modified_time = last_modified_time
        
    @staticmethod
    def get_message_addresses(message):
        """Get the origin and destination addresses of a message.

        :param message: A message object supporting get() to access its
            headers, like email.Message or bzrlib.email_message.EmailMessage.
        :return: A pair (from_email, to_emails), where from_email is the email
            address in the From header, and to_emails a list of all the
            addresses in the To, Cc, and Bcc headers.
        """
        from_email = Utils.parseaddr(message.get('From', None))[1]
        to_full_addresses = []
        for header in ['To', 'Cc', 'Bcc']:
            value = message.get(header, None)
            if value:
                to_full_addresses.append(value)
        to_emails = [ pair[1] for pair in
                Utils.getaddresses(to_full_addresses) ]

        return from_email, to_emails

    def send_email(self, message):
        """Send an email message.

        :param message: An email.Message or email.MIMEMultipart object.
        :return: None
        """
        from_email, to_emails = self.get_message_addresses(message)
        
        try:
            self._connect()
            self._connection.sendmail(from_email, to_emails,
                                      message.as_string())
        except (socket.error, smtplib.SMTPException), e:
            self.disconnect()
            raise SendEmailError(e)
        
    def _connect(self):
        """If we haven't connected, connect and authenticate."""
        if self._connection is not None:
            return

        self._create_connection()
        self._authenticate()
    
    def _create_connection(self):
        """Create an SMTP connection."""
        self._connection = smtplib.SMTP()
        self._connection.connect(self._smtp_server)

        # If this fails, it just returns an error, but it shouldn't raise an
        # exception unless something goes really wrong (in which case we want
        # to fail anyway).
        self._connection.starttls()

    def _authenticate(self):
        """If necessary authenticate yourself to the server."""
        if self._smtp_username is None:
            return

        self._connection.login(self._smtp_username, self._smtp_password)
        
    def disconnect(self):
        "Disconnect from the SMTP server"
         
        if self._connection is not None:
            try:
                self._connection.quit()
            except (socket.error, smtplib.SMTPException), e:
                print >>sys.stderr, 'Error closing socket: %s' % (e)
                
            self._connection = None
        
    
class EmailConfiguration:
    # Defines settings to use for sending an email.
    # This consists of Cheetah template definitions for the subject, from, to,
    # and body of the email.
    
    def __init__(self, config):
        self._config = config
        self._body_template_file_last_modified = 0
        
    def load(self, config, section_name):
        self._to_template = Template.compile(source=config.get(section_name, 'to'), baseclass=dict)
        self._from_template = Template.compile(source=config.get(section_name, 'from'), baseclass=dict)
        self._subject_template = Template.compile(config.get(section_name, 'subject'), baseclass=dict)
        self._body_template_file = get_option(config, section_name, 'body', None)
        self._load_body_template_file()
        
    def _load_body_template_file(self):
        if self._body_template_file is None:
            self._body_template = None
            self._is_html = False
            self._body_template_file_last_modified = 0
        else:
            self._body_template_file_last_modified = os.path.getmtime(self._body_template_file)
            self._body_template = Template.compile(file=self._body_template_file, baseclass=dict)
            self._is_html = self._body_template_file.endswith('.html')
        
    def compose_email(self, data):
        # Composes an email message using these settings and the supplied data.
        # data is a dictionary mapping names to values for substitution in the
        # templates.
        
        actual_data = dict(data)
        
        for key, value in self._config.template_settings:
            actual_data[key] = value
            
        to_address = unicode(self._to_template(actual_data))
        from_address = unicode(self._from_template(actual_data))
        subject = unicode(self._subject_template(actual_data))
        
        if self._body_template is None:
            body = None
        else:
            body = unicode(self._body_template(actual_data))
            
        message = EmailMessage(from_address, to_address, subject, None if self._is_html else body)
        
        if not body is None and self._is_html:
            # Using HTML for message. Add it as an inline attachment.
            message.add_inline_attachment(body, None, 'html')
        
        return message

        
class SendEmailError(Exception):
    # Exception raised when there is an error sending an email.

    def __init__(self, error):
        self.error = error
        
    def __str__(self):
        return 'Error sending email: ' + str(self.error)
        
        
def get_option(config, section, name, default):
    # Gets an option from the given ConfigParser and returns the default value
    # if it doesn't exist.
    try:
        return config.get(section, name) 
    except ConfigParser.NoOptionError:
        return default
        
def get_int_option(config, section, name, default):
    # Gets an option from the given ConfigParser and returns the default vaule
    # if it doesn't exist.
    try:
        return config.getint(section, name) 
    except ConfigParser.NoOptionError:
        return default
        
def get_bool_option(config, section, name, default):
    # Gets an option from the given ConfigParser and returns the default vaule
    # if it doesn't exist.
    try:
        return config.getboolean(section, name) 
    except ConfigParser.NoOptionError:
        return default

